from __future__ import annotations

from typing import Any

import torch
from torch import Tensor

from ._feature import Feature


class RotationVector(Feature):
    @classmethod
    def _wrap(cls, tensor: Tensor) -> RotationVector:
        return tensor.as_subclass(cls)

    def __new__(
        cls,
        data: Any,
        *,
        dtype: torch.dtype | None = None,
        device: torch.device | str | int | None = None,
        requires_grad: bool | None = None,
    ) -> RotationVector:
        tensor = cls._to_tensor(
            data,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )

        if tensor.ndim == 0:
            raise ValueError

        if tensor.shape[-1] != 3:
            raise ValueError

        if tensor.ndim == 1:
            tensor = torch.unsqueeze(tensor, 0)

        return cls._wrap(tensor)

    @classmethod
    def wrap_like(cls, other: RotationVector, tensor: Tensor) -> RotationVector:
        return cls._wrap(tensor)

    def __repr__(self, *, tensor_contents: Any = None) -> str:
        return self._make_repr()


_RotationVectorType = Tensor | RotationVector
_RotationVectorTypeJIT = Tensor

_TensorRotationVectorType = Tensor | RotationVector
_TensorRotationVectorTypeJIT = Tensor



import torch
from torch import Tensor


def differentiate_chebyshev_polynomial(
    input: Tensor,
    order=1,
    scale=1,
    axis=0,
) -> Tensor:
    r"""
    Returns the derivative of a polynomial.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    order : Tensor, optional

    scale : Tensor, optional

    dim : int, default=0

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the derivative.
    """
    if order < 0:
        raise ValueError

    input = torch.atleast_1d(input)

    if order == 0:
        return input

    output = torch.moveaxis(input, axis, 0)

    n = output.shape[0]

    if order >= n:
        output = torch.zeros_like(output[:1])
    else:
        for _ in range(order):
            n = n - 1

            output = output * scale

            derivative = torch.empty((n,) + output.shape[1:], dtype=output.dtype)

            for i in range(0, n - 2):
                j = n - i

                derivative[j - 1] = (2 * j) * output[j]

                output[j - 2] += (j * output[j]) / (j - 2)

            if n > 1:
                derivative[1] = 4 * output[2]

            derivative[0] = output[1]

    return torch.moveaxis(output, 0, axis)



import beignet
import pytest
import torch


def test_legendre_polynomial_companion():
    with pytest.raises(ValueError):
        beignet.legendre_polynomial_companion(torch.tensor([]))

    with pytest.raises(ValueError):
        beignet.legendre_polynomial_companion(torch.tensor([1]))

    for index in range(1, 5):
        output = beignet.legendre_polynomial_companion(
            torch.tensor([0.0] * index + [1.0]),
        )

        assert output.shape == (index, index)

    assert beignet.legendre_polynomial_companion(torch.tensor([1, 2]))[0, 0] == -0.5



import beignet
import pytest
import torch


def test_polynomial_companion():
    with pytest.raises(ValueError):
        beignet.polynomial_companion(torch.tensor([]))

    with pytest.raises(ValueError):
        beignet.polynomial_companion(torch.tensor([1]))

    for i in range(1, 5):
        output = beignet.polynomial_companion(
            torch.tensor([0.0] * i + [1.0]),
        )

        assert output.shape == (i, i)

    output = beignet.polynomial_companion(
        torch.tensor([1, 2]),
    )

    assert output[0, 0] == -0.5



from typing import Literal

import torch
from torch import Tensor


def multiply_probabilists_hermite_polynomial_by_x(
    input: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros(input.shape[0] + 1, dtype=input.dtype)

    output[1] = input[0]

    index = torch.arange(1, input.shape[0])

    output[index + 1] = input[index]
    output[index - 1] = output[index - 1] + input[index] * index

    if mode == "same":
        output = output[: input.shape[0]]

    return output



import sys
import unittest.mock
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
from beignet.datasets import UniProtDataset


class TestUniProtDataset:
    @pytest.mark.skipif(sys.platform == "win32", reason="windows")
    def test___init__(self):
        with NamedTemporaryFile() as descriptor:
            with unittest.mock.patch(
                "pooch.retrieve",
                lambda a, b, c, d, **_: descriptor.name,
            ):
                dataset = UniProtDataset(
                    "https://example.com",
                    descriptor.name,
                )

                assert dataset.root == Path(descriptor.name).resolve()

                assert dataset.transform is None

                assert dataset.target_transform is None



import math

import beignet
import torch


# def test_polynomial_from_roots():
#     coefficients = [
#         torch.tensor([1.0]),
#         torch.tensor([0.0, 1]),
#         torch.tensor([-1.0, 0, 2]),
#         torch.tensor([0.0, -3, 0, 4]),
#         torch.tensor([1.0, 0, -8, 0, 8]),
#         torch.tensor([0.0, 5, 0, -20, 0, 16]),
#         torch.tensor([-1.0, 0, 18, 0, -48, 0, 32]),
#         torch.tensor([0.0, -7, 0, 56, 0, -112, 0, 64]),
#         torch.tensor([1.0, 0, -32, 0, 160, 0, -256, 0, 128]),
#         torch.tensor([0.0, 9, 0, -120, 0, 432, 0, -576, 0, 256]),
#     ]
#
#     torch.testing.assert_close(
#         beignet.trim_polynomial_coefficients(
#             beignet.polynomial_from_roots(
#                 torch.tensor([]),
#             ),
#             tol=0.000001,
#         ),
#         torch.tensor([1.0]),
#     )
#
#     for index in range(1, 5):
#         input = torch.linspace(-math.pi, 0.0, 2 * index + 1)
#
#         input = input[1::2]
#
#         input = torch.cos(input)
#
#         output = beignet.polynomial_from_roots(input) * 2 ** (index - 1)
#
#         torch.testing.assert_close(
#             beignet.trim_polynomial_coefficients(
#                 output,
#                 tol=0.000001,
#             ),
#             beignet.trim_polynomial_coefficients(
#                 coefficients[index],
#                 tol=0.000001,
#             ),
#         )



from os import PathLike
from pathlib import Path
from typing import Callable, Tuple, TypeVar

import numpy
import tqdm

from beignet.io import ThreadSafeFile

from ..transforms import Transform
from ._sized_sequence_dataset import SizedSequenceDataset

T = TypeVar("T")


class FASTADataset(SizedSequenceDataset):
    def __init__(
        self,
        root: str | PathLike,
        *,
        transform: Callable | Transform | None = None,
    ):
        if isinstance(root, str):
            self.root = Path(root)

        self.root = self.root.resolve()

        if not self.root.exists():
            raise FileNotFoundError

        self.data = ThreadSafeFile(self.root, open)

        offsets = Path(f"{self.root}.offsets.npy")

        if offsets.exists():
            self.offsets, sizes = numpy.load(f"{offsets}")
        else:
            self.offsets, sizes = self._build_index()

            numpy.save(f"{offsets}", numpy.stack([self.offsets, sizes]))

        self.transform = transform

        super().__init__(self.root, sizes)

    def __getitem__(self, index: int) -> Tuple[str, str]:
        x = self.get(index)

        if self.transform:
            x = self.transform(x)

        return x

    def __len__(self) -> int:
        return self.offsets.size

    def get(self, index: int) -> (str, str):
        self.data.seek(self.offsets[index])

        if index == len(self) - 1:
            data = self.data.read()
        else:
            data = self.data.read(self.offsets[index + 1] - self.offsets[index])

        description, *sequence = data.split("\n")

        return "".join(sequence), description

    def _build_index(self) -> (numpy.ndarray, numpy.ndarray):
        with open(self.root, "r") as file:
            content = file.read()

        offsets, sizes = [], []

        current_offset, current_size = 0, 0

        parsing = False

        for sequence in tqdm.tqdm(content.splitlines(keepends=True)):
            characters = len(sequence)

            if sequence.startswith(">"):
                if parsing:
                    sizes = [*sizes, current_size]

                    current_size = 0

                offsets = [*offsets, current_offset]

                parsing = True
            elif parsing:
                current_size = current_size + len(sequence.rstrip("\n"))

            current_offset = current_offset + characters

        if parsing:
            sizes = [*sizes, current_size]

        offsets = numpy.array(offsets, dtype=numpy.int64)

        sizes = numpy.array(sizes, dtype=numpy.int64)

        return offsets, sizes



# import functools
# from typing import Callable
#
# import beignet.func
# import hypothesis
# import hypothesis.strategies
# import torch.testing
# from torch import Tensor
#
#
# def map_product(fn: Callable) -> Callable:
#     return torch.vmap(
#         torch.vmap(
#             fn,
#             in_dims=(0, None),
#             out_dims=0,
#         ),
#         in_dims=(None, 0),
#         out_dims=0,
#     )
#
#
# @hypothesis.strategies.composite
# def _strategy(function):
#     dtype = function(
#         hypothesis.strategies.sampled_from(
#             [
#                 torch.float32,
#                 torch.float64,
#             ],
#         ),
#     )
#
#     maximum_size = function(
#         hypothesis.strategies.floats(
#             min_value=1.0,
#             max_value=8.0,
#         ),
#     )
#
#     particles = function(
#         hypothesis.strategies.integers(
#             min_value=16,
#             max_value=32,
#         ),
#     )
#
#     spatial_dimension = function(
#         hypothesis.strategies.integers(
#             min_value=1,
#             max_value=3,
#         ),
#     )
#
#     return (
#         dtype,
#         torch.rand([particles, spatial_dimension], dtype=dtype),
#         particles,
#         torch.rand([spatial_dimension], dtype=dtype) * maximum_size,
#         spatial_dimension,
#     )
#
#
# @hypothesis.given(_strategy())
# @hypothesis.settings(deadline=None)
# def test_space(data):
#     dtype, input, particles, size, spatial_dimension = data
#
#     displacement_fn, shift_fn = beignet.func.space(size, parallelepiped=False)
#
#     (
#         parallelepiped_displacement_fn,
#         parallelepiped_shift_fn,
#     ) = beignet.func.space(
#         torch.diag(size),
#     )
#
#     standardized_input = input * size
#
#     displacement_fn = map_product(displacement_fn)
#
#     parallelepiped_displacement_fn = map_product(
#         parallelepiped_displacement_fn,
#     )
#
#     torch.testing.assert_close(
#         displacement_fn(
#             standardized_input,
#             standardized_input,
#         ),
#         parallelepiped_displacement_fn(
#             input,
#             input,
#         ),
#     )
#
#     displacement = torch.randn([particles, spatial_dimension], dtype=dtype)
#
#     torch.testing.assert_close(
#         shift_fn(standardized_input, displacement),
#         parallelepiped_shift_fn(input, displacement) * size,
#     )
#
#     def f(input: Tensor) -> Tensor:
#         return torch.sum(displacement_fn(input, input) ** 2)
#
#     def g(input: Tensor) -> Tensor:
#         return torch.sum(parallelepiped_displacement_fn(input, input) ** 2)
#
#     torch.testing.assert_close(
#         torch.func.grad(f)(standardized_input),
#         torch.func.grad(g)(input),
#         rtol=0.0001,
#         atol=0.0001,
#     )
#
#     size_a = 10.0 * torch.rand([])
#     size_b = 10.0 * torch.rand([], dtype=dtype)
#
#     transform_a = 0.5 * torch.randn([spatial_dimension, spatial_dimension])
#     transform_b = 0.5 * torch.randn(
#         [spatial_dimension, spatial_dimension],
#         dtype=dtype,
#     )
#
#     transform_a = size_a * (torch.eye(spatial_dimension) + transform_a)
#     transform_b = size_b * (torch.eye(spatial_dimension) + transform_b)
#
#     displacement_fn_a, shift_fn_a = beignet.func.space(transform_a)
#     displacement_fn_b, shift_fn_b = beignet.func.space(transform_b)
#
#     displacement = torch.randn([particles, spatial_dimension], dtype=dtype)
#
#     torch.testing.assert_close(
#         map_product(
#             functools.partial(
#                 displacement_fn_a,
#                 transform=transform_b,
#             ),
#         )(input, input),
#         map_product(displacement_fn_b)(input, input),
#     )
#
#     torch.testing.assert_close(
#         shift_fn_a(input, displacement, transform=transform_b),
#         shift_fn_b(input, displacement),
#     )



import beignet
import torch


def test_multiply_physicists_hermite_polynomial_by_x():
    torch.testing.assert_close(
        beignet.trim_physicists_hermite_polynomial_coefficients(
            beignet.multiply_physicists_hermite_polynomial_by_x(
                torch.tensor([0.0]),
            ),
            tol=0.000001,
        ),
        torch.tensor([0.0]),
    )

    torch.testing.assert_close(
        beignet.multiply_physicists_hermite_polynomial_by_x(
            torch.tensor([1.0]),
        ),
        torch.tensor([0.0, 0.5]),
    )

    for i in range(1, 5):
        torch.testing.assert_close(
            beignet.multiply_physicists_hermite_polynomial_by_x(
                torch.tensor([0.0] * i + [1.0]),
            ),
            torch.tensor([0.0] * (i - 1) + [i, 0.0, 0.5]),
        )



import beignet
import pytest
import torch


def test_chebyshev_extrema():
    with pytest.raises(ValueError):
        beignet.chebyshev_extrema(1.5)

    with pytest.raises(ValueError):
        beignet.chebyshev_extrema(1)

    torch.testing.assert_close(
        beignet.chebyshev_extrema(2),
        torch.tensor([-1.0, 1.0]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_extrema(3),
        torch.tensor([-1.0, 0.0, 1.0]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_extrema(4),
        torch.tensor([-1.0, -0.5, 0.5, 1.0]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_extrema(5),
        torch.tensor([-1.0, -0.707106781187, 0, 0.707106781187, 1.0]),
    )



import torch
from torch import Tensor


def quaternion_magnitude(input: Tensor, canonical=False) -> Tensor:
    r"""
    Rotation quaternion magnitudes.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation quaternions.

    Returns
    -------
    output : Tensor, shape=(...)
        Angles in radians. Magnitudes will be in the range :math:`[0, \pi]`.
    """
    output = torch.empty(
        input.shape[0],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
        requires_grad=input.requires_grad,
    )

    for j in range(input.shape[0]):
        a = input[j, 0]
        b = input[j, 1]
        c = input[j, 2]
        d = input[j, 3]

        x = torch.atan2(torch.sqrt(a**2 + b**2 + c**2), torch.abs(d))

        output[j] = x * 2.0

    return output



import beignet
import torch


def test_linear_polynomial():
    torch.testing.assert_close(
        beignet.linear_polynomial(3.0, 4.0),
        torch.tensor([3.0, 4.0]),
    )

    torch.testing.assert_close(
        beignet.linear_polynomial(3.0, 0.0),
        torch.tensor([3.0, 0.0]),
    )



from torch import Tensor

from ._error_erf import error_erf


def error_erfi(input: Tensor, *, out: Tensor | None = None) -> Tensor:
    r"""
    Imaginary error function.

    Parameters
    ----------
    input : Tensor
        Input tensor.

    out : Tensor, optional
        Output tensor.

    Returns
    -------
    Tensor
    """
    output = -1.0j * error_erf(1.0j * input)

    if out is not None:
        out.copy_(output)

        return out

    return output



from os import PathLike
from typing import Any, Callable

import mdtraj
from mdtraj import Trajectory

from ._trajectory_dataset import TrajectoryDataset


class PDBTrajectoryDataset(TrajectoryDataset):
    def __init__(
        self,
        root: str | PathLike,
        transform: Callable[[Trajectory], Any] | None = None,
        stride: int | None = None,
        **kwargs,
    ):
        super().__init__(
            func=mdtraj.load_pdb,
            extension="pdb",
            root=root,
            transform=transform,
            stride=stride,
            **kwargs,
        )



import torch
from torch import Tensor

from ._add_chebyshev_polynomial import add_chebyshev_polynomial
from ._multiply_chebyshev_polynomial_by_x import multiply_chebyshev_polynomial_by_x


def polynomial_to_chebyshev_polynomial(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros_like(input)

    for i in range(0, input.shape[0] - 1 + 1):
        output = multiply_chebyshev_polynomial_by_x(output, mode="same")

        output = add_chebyshev_polynomial(output, input[input.shape[0] - 1 - i])

    return output



import torch
from torch import Tensor
from torch.autograd import Function


def _apply_transform(input: Tensor, transform: Tensor) -> Tensor:
    """
    Applies an affine transformation to the position vector.

    Parameters
    ----------
    input : Tensor
        Position, must have the shape `(..., dimension)`.

    transform : Tensor
        The affine transformation matrix, must be a scalar, a vector, or a
        matrix with the shape `(dimension, dimension)`.

    Returns
    -------
    Tensor
        Affine transformed position vector, has the same shape as the
        position vector.
    """
    match transform.ndim:
        case 0:
            return input * transform
        case 1:
            return torch.einsum("i,...i->...i", transform, input)
        case 2:
            return torch.einsum("ij,...j->...i", transform, input)
        case _:
            raise ValueError


class _ApplyTransform(Function):
    generate_vmap_rule = True

    @staticmethod
    def forward(transform: Tensor, input: Tensor) -> Tensor:
        """
        Return affine transformed position.

        Parameters
        ----------
        transform : Tensor
            Affine transformation matrix, must have shape
            `(dimension, dimension)`.

        input : Tensor
            Position, must have shape `(..., dimension)`.

        Returns
        -------
        Tensor
            Affine transformed position of shape `(..., dimension)`.
        """
        return _apply_transform(input, transform)

    @staticmethod
    def setup_context(ctx, inputs, output):
        transform, input = inputs

        ctx.save_for_backward(transform, input, output)

    @staticmethod
    def jvp(ctx, grad_transform: Tensor, grad_input: Tensor) -> (Tensor, Tensor):
        transform, input, _ = ctx.saved_tensors

        output = _apply_transform(input, transform)

        grad_output = grad_input + _apply_transform(input, grad_transform)

        return output, grad_output

    @staticmethod
    def backward(ctx, grad_output: Tensor) -> (Tensor, Tensor):
        _, _, output = ctx.saved_tensors

        return output, grad_output


def apply_transform(input: Tensor, transform: Tensor) -> Tensor:
    """
    Return affine transformed position.

    Parameters
    ----------
    input : Tensor
        Position, must have shape `(..., dimension)`.

    transform : Tensor
        Affine transformation matrix, must have shape
        `(dimension, dimension)`.

    Returns
    -------
    Tensor
        Affine transformed position of shape `(..., dimension)`.
    """
    return _ApplyTransform.apply(transform, input)



from os import PathLike

from numpy.typing import ArrayLike

from ._sequence_dataset import SequenceDataset


class SizedSequenceDataset(SequenceDataset):
    def __init__(
        self,
        root: str | PathLike,
        sizes: ArrayLike,
        *args,
        **kwargs,
    ):
        super().__init__(root, *args, **kwargs)

        self.sizes = sizes

    def __len__(self) -> int:
        return len(self.sizes)



import beignet
import torch


def test_divide_legendre_polynomial():
    for i in range(5):
        for j in range(5):
            quotient, remainder = beignet.divide_legendre_polynomial(
                beignet.add_legendre_polynomial(
                    torch.tensor([0.0] * i + [1.0]),
                    torch.tensor([0.0] * j + [1.0]),
                ),
                torch.tensor([0.0] * i + [1.0]),
            )

            torch.testing.assert_close(
                beignet.trim_legendre_polynomial_coefficients(
                    beignet.add_legendre_polynomial(
                        beignet.multiply_legendre_polynomial(
                            quotient,
                            torch.tensor([0.0] * i + [1.0]),
                        ),
                        remainder,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_legendre_polynomial_coefficients(
                    beignet.add_legendre_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        torch.tensor([0.0] * j + [1.0]),
                    ),
                    tol=0.000001,
                ),
            )



from typing import Literal

import torch
from torch import Tensor


def multiply_chebyshev_polynomial_by_x(
    input: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros(input.shape[0] + 1, dtype=input.dtype)

    output[1] = input[0]

    if input.shape[0] > 1:
        output[2:] = input[1:] / 2

        output[0:-2] = output[0:-2] + input[1:] / 2

    if mode == "same":
        output = output[: input.shape[0]]

    return output



from os import PathLike
from typing import Callable

from beignet.transforms import Transform

from ._uniprot_dataset import UniProtDataset


class SwissProtDataset(UniProtDataset):
    def __init__(
        self,
        root: str | PathLike | None = None,
        *,
        transform: Callable | Transform | None = None,
        target_transform: Callable | Transform | None = None,
    ):
        """
        Parameters
        ----------
        root : str | PathLike, optional
            Root directory where the dataset subdirectory exists or, if
            `download` is `True`, the directory where the dataset subdirectory
            will be created and the dataset downloaded.

        transform : Callable, optional
            A `Callable` or `Transform` that that maps a sequence to a
            transformed sequence (default: `None`).

        target_transform : Callable, optional
            A `Callable` or `Transform` that maps a target to a transformed
            target (default: `None`).
        """
        super().__init__(
            "https://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.fasta.gz",
            root,
            "md5:0766df3e5785fc5f1cfc496aa89e86ad",
            transform=transform,
            target_transform=target_transform,
        )



from pathlib import Path
from typing import Callable, Dict, Tuple, Union

import torch
from pandas import DataFrame
from torch import Tensor

from beignet.transforms import Transform

from ._atom3d_dataset import ATOM3DDataset


class ATOM3DPSRDataset(ATOM3DDataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ):
        super().__init__(
            root,
            "raw/casp5_to_13/data",
            "https://zenodo.org/record/4915648/files/PSR-raw.tar.gz",
            "PSR",
            checksum="80caef3c98febb70951fa244c8303039",
            download=download,
        )

        self._transform_fn = transform

        self._target_transform_fn = target_transform

    def __getitem__(
        self,
        index: int,
    ) -> Tuple[DataFrame, Dict[str, Tensor]]:
        item = super().__getitem__(index)

        features = DataFrame(**item["atoms"])

        if self._transform_fn is not None:
            features = self._transform_fn(features)

        target = item["scores"]

        for k, _ in target.items():
            target[k] = torch.tensor(target[k])

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(target)

        return features, target



from typing import Literal

import torch
from torch import Tensor


def multiply_laguerre_polynomial_by_x(
    input: Tensor,
    mode: Literal["full", "same", "valid"] = "full",
) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros(input.shape[0] + 1, dtype=input.dtype)

    output[0] = +input[0]
    output[1] = -input[0]

    i = torch.arange(1, input.shape[0])

    output[i + 1] = -input[i] * (i + 1)

    output[i] = output[i] + input[i] * (2 * i + 1)

    output[i - 1] = output[i - 1] - input[i] * i

    if mode == "same":
        output = output[: input.shape[0]]

    return output



import torch
from torch import Tensor

from ._add_polynomial import add_polynomial
from ._multiply_polynomial_by_x import multiply_polynomial_by_x
from ._subtract_polynomial import subtract_polynomial


def physicists_hermite_polynomial_to_polynomial(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    n = input.shape[0]

    if n == 1:
        return input

    if n == 2:
        input[1] = input[1] * 2

        return input
    else:
        c0 = torch.zeros_like(input)
        c0[0] = input[-2]

        c1 = torch.zeros_like(input)
        c1[0] = input[-1]

        def body(k, c0c1):
            i = n - 1 - k
            c0, c1 = c0c1
            tmp = c0
            c0 = subtract_polynomial(input[i - 2], c1 * (2 * (i - 1)))
            c1 = add_polynomial(tmp, multiply_polynomial_by_x(c1, "same") * 2)
            return c0, c1

        x = (c0, c1)

        y = x

        for index in range(0, n - 2):
            y = body(index, y)

        c0, c1 = y

        return add_polynomial(c0, multiply_polynomial_by_x(c1, "same") * 2)



from pathlib import Path
from typing import Callable, Dict, Tuple, Union

import torch
from pandas import DataFrame
from torch import Tensor

from beignet.transforms import Transform

from ._atom3d_dataset import ATOM3DDataset


class ATOM3DSMPDataset(ATOM3DDataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ):
        """
        ATOM3D Small Molecule Properties (SMP) is a dataset of structures
        and energetic, electronic, and thermodynamic properties for 134,000
        stable small organic molecules, obtained from quantum-chemical
        calculations. The task is to predict the molecular properties from
        the ground-state structure.

        Some molecules have been excluded because they failed consistency
        tests or were not properly processed.

        Each sample is a pair of features and a target, where features is
        the molecule’s atomic coordinates and target is a dictionary of the
        following energetic, electronic, and thermodynamic properties:

        .. list-table:: Target
           :widths: 20 20 60
           :header-rows: 1

           * - Key
             - Unit
             - Description
           * - a
             - GHz
             - Rotational constant A
           * - b
             - GHz
             - Rotational constant B
           * - c
             - GHz
             - Rotational constant C
           * - mu
             - Debye
             - Dipole moment
           * - alpha
             - Bohr^3
             - Isotropic polarizability
           * - homo
             - Hartree
             - Energy of Highest occupied molecular orbital (HOMO)
           * - lumo
             - Hartree
             - Energy of Lowest occupied molecular orbital (LUMO)
           * - gap
             - Hartree
             - Gap, difference between LUMO and HOMO
           * - r2
             - Bohr^2
             - Electronic spatial extent
           * - zpve
             - Hartree
             - Zero point vibrational energy
           * - u0
             - Hartree
             - Internal energy at 0 K
           * - u
             - Hartree
             - Internal energy at 298.15 K
           * - h
             - Hartree
             - Enthalpy at 298.15 K
           * - g
             - Hartree
             - Free energy at 298.15 K
           * - cv
             - cal/(mol K)
             - Heat capacity at 298.15 K

        Parameters
        ----------
        root : Union[str, Path]
            The root directory of the dataset.

        download : bool, optional
            If True, download the dataset from the specified source,
            by default `False`.

        transform : Union[Callable, Transform, None], optional
            The transformation function to be applied to the features,
            by default `None`.

        target_transform : Union[Callable, Transform, None], optional
            The transformation function to be applied to the targets,
            by default `None`.
        """
        super().__init__(
            root,
            "raw/QM9/data",
            "https://zenodo.org/record/4911142/files/SMP-raw.tar.gz",
            "SMP",
            checksum="52cc7955c0f80f7dd9faf041e171f405",
            download=download,
        )

        self._transform_fn = transform

        self._target_transform_fn = target_transform

    def __getitem__(
        self,
        index: int,
    ) -> Tuple[Tuple[DataFrame], Dict[str, Tensor]]:
        """
        Parameters
        ----------
        index : int
            The index of the item to retrieve from the dataset.

        Returns
        -------
        Tuple[Tuple[DataFrame], Dict[str, Tensor]]
            A tuple containing the features and target of the item.

        """
        item = super().__getitem__(index)

        features = DataFrame(**item["atoms"])

        if self._transform_fn is not None:
            features = self._transform_fn(features)

        target = {}

        for k, v in zip(
            [
                "a",
                "b",
                "c",
                "mu",
                "alpha",
                "homo",
                "lumo",
                "gap",
                "r2",
                "zpve",
                "u0",
                "u",
                "h",
                "g",
                "cv",
            ],
            item["labels"],
            strict=False,
        ):
            target[k] = torch.tensor(v)

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(target)

        return features, target



import pytest
import torch
from beignet.features import RotationMatrix
from torch import Size


class TestRotationMatrix:
    @pytest.fixture
    def rotation_matrix(self) -> RotationMatrix:
        return RotationMatrix(
            torch.tensor(
                [
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                ],
            ),
        )

    def test___new__(self):
        result = RotationMatrix(
            torch.tensor(
                [
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                ],
            ),
        )

        assert isinstance(result, RotationMatrix)

        assert result.shape == Size([1, 3, 3])

        with pytest.raises(ValueError):
            RotationMatrix(torch.tensor([1, 2, 3]))

    def test___repr__(self, rotation_matrix: RotationMatrix):
        assert isinstance(rotation_matrix.__repr__(), str)

    def test__wrap(self):
        result = RotationMatrix._wrap(
            torch.tensor(
                [
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                ],
            ),
        )

        assert isinstance(result, RotationMatrix)

        assert result.shape == Size([1, 3, 3])

    def test_wrap_like(self, rotation_matrix: RotationMatrix):
        result = RotationMatrix.wrap_like(
            rotation_matrix,
            torch.tensor(
                [
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                ],
            ),
        )

        assert isinstance(result, RotationMatrix)

        assert result.shape == Size([1, 3, 3])



from ._euler_angle import EulerAngle
from ._feature import Feature
from ._quaternion import Quaternion
from ._rotation_matrix import RotationMatrix
from ._rotation_vector import RotationVector

__all__ = [
    "EulerAngle",
    "Feature",
    "Quaternion",
    "RotationMatrix",
    "RotationVector",
]



# import beignet
# import hypothesis.strategies
# import numpy
# import torch
# from scipy.spatial.transform import Rotation, Slerp
#
# def test_slerp():
#     # t = 0
#     torch.testing.assert_close(
#         beignet.quaternion_slerp(
#             torch.tensor([+0.00000]),
#             torch.tensor([+0.00000, +1.00000]),
#             torch.tensor(
#                 [
#                     [+1.00000, +0.00000, +0.00000, +0.00000],
#                     [+0.00000, +1.00000, +0.00000, +0.00000],
#                 ]
#             ),
#         ),
#         torch.tensor([[+1.00000, +0.00000, +0.00000, +0.00000]]),
#     )
#
#     # t = 1
#     torch.testing.assert_close(
#         beignet.quaternion_slerp(
#             torch.tensor([+1.00000]),
#             torch.tensor([+0.00000, +1.00000]),
#             torch.tensor(
#                 [
#                     [+1.00000, +0.00000, +0.00000, +0.00000],
#                     [+0.00000, +1.00000, +0.00000, +0.00000],
#                 ]
#             ),
#         ),
#         torch.tensor([[+0.00000, +1.00000, +0.00000, +0.00000]]),
#     )
#
#     # SMALL (ACUTE) ANGLE BETWEEN QUATERNIONS
#     torch.testing.assert_close(
#         beignet.quaternion_slerp(
#             torch.tensor([+0.50000]),
#             torch.tensor([+0.00000, +1.00000]),
#             torch.tensor(
#                 [
#                     [+1.00000, +0.00000, +0.00000, +0.00000],
#                     [+0.70710, +0.70710, +0.00000, +0.00000],
#                 ],
#             ),
#         ),
#         torch.reshape(
#             torch.tensor([+0.92388, +0.38268, +0.00000, +0.00000]),
#             [1, -1],
#         ),
#     )
#
#     # LARGE (OBTUSE) ANGLE BETWEEN QUATERNIONS
#     torch.testing.assert_close(
#         beignet.quaternion_slerp(
#             torch.tensor([+0.50000]),
#             torch.tensor([+0.00000, +1.00000]),
#             torch.tensor(
#                 [
#                     [+1.00000, +0.00000, +0.00000, +0.00000],
#                     [-1.00000, +0.00000, +0.00000, +0.00000],
#                 ]
#             ),
#         ),
#         torch.reshape(
#             torch.tensor([+1.00000, +0.00000, +0.00000, +0.00000]),
#             [1, -1],
#         ),
#     )
#
#
# @hypothesis.strategies.composite
# def slerp_parameters(f):
#     n = f(
#         hypothesis.strategies.integers(
#             min_value=2,
#             max_value=8,
#         ),
#     )
#
#     times = numpy.sort(
#         f(
#             hypothesis.strategies.lists(
#                 hypothesis.strategies.floats(
#                     allow_infinity=False,
#                     allow_nan=False,
#                 ),
#                 min_size=n,
#                 max_size=n,
#                 unique=True,
#             ),
#         ),
#     )
#
#     min_value = numpy.min(times)
#     max_value = numpy.max(times)
#
#     input = numpy.sort(
#         f(
#             hypothesis.strategies.lists(
#                 hypothesis.strategies.floats(
#                     min_value=min_value,
#                     max_value=max_value,
#                 ),
#                 min_size=1,
#                 max_size=8,
#                 unique=True,
#             ),
#         ),
#     )
#
#     rotations = f(
#         hypothesis.strategies.lists(
#             hypothesis.strategies.lists(
#                 hypothesis.strategies.floats(
#                     numpy.finfo(numpy.float32).eps,
#                     1.0,
#                 ),
#                 min_size=4,
#                 max_size=4,
#             ),
#             min_size=n,
#             max_size=n,
#         ),
#     )
#
#     rotations = Rotation.from_quat(rotations)
#
#     return [
#         [
#             torch.from_numpy(input),
#             torch.from_numpy(times),
#             torch.from_numpy(rotations.as_quat(canonical=True)),
#         ],
#         torch.from_numpy(
#             Slerp(times, rotations)(input).as_quat(canonical=True),
#         ),
#     ]
#
#
# @hypothesis.given(slerp_parameters())
# def test_slerp_properties(data):
#     parameters, expected_rotations = data
#
#     torch.testing.assert_close(
#         beignet.quaternion_slerp(*parameters), expected_rotations
#     )



from torch import Tensor

from ._quaternion_mean import quaternion_mean
from ._quaternion_to_rotation_vector import quaternion_to_rotation_vector
from ._rotation_vector_to_quaternion import rotation_vector_to_quaternion


def rotation_vector_mean(
    input: Tensor,
    weight: Tensor | None = None,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Compose rotation vectors.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation vectors.

    weight : Tensor, shape=(..., 4), optional
        Relative importance of rotation matrices.

    degrees : bool, optional
        If `True`, rotation vector magnitudes are assumed to be in degrees.
        Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 4)
        Rotation vectors mean.
    """
    return quaternion_to_rotation_vector(
        quaternion_mean(
            rotation_vector_to_quaternion(
                input,
                degrees,
            ),
            weight,
        ),
        degrees,
    )



import torch
from torch import Tensor


def laguerre_polynomial_companion(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    if input.shape[0] < 2:
        raise ValueError

    if input.shape[0] == 2:
        return torch.tensor([[1 + input[0] / input[1]]])

    n = input.shape[0] - 1

    output = torch.reshape(torch.zeros([n, n], dtype=input.dtype), [-1])

    output[1 :: n + 1] = -torch.arange(1, n)

    output[0 :: n + 1] = 2.0 * torch.arange(n) + 1.0

    output[n :: n + 1] = -torch.arange(1, n)

    output = torch.reshape(output, [n, n])

    output[:, -1] += (input[:-1] / input[-1]) * n

    return output



from torch import Tensor

from ._evaluate_polynomial import evaluate_polynomial


def evaluate_polynomial_2d(x: Tensor, y: Tensor, coefficients: Tensor) -> Tensor:
    r"""
    Parameters
    ----------
    x : Tensor

    y : Tensor

    coefficients : Tensor

    Returns
    -------
    output : Tensor
    """
    points = [x, y]

    if not all(a.shape == points[0].shape for a in points[1:]):
        match len(points):
            case 2:
                raise ValueError
            case 3:
                raise ValueError
            case _:
                raise ValueError

    points = iter(points)

    output = evaluate_polynomial(
        next(points),
        coefficients,
    )

    for x in points:
        output = evaluate_polynomial(
            x,
            output,
            tensor=False,
        )

    return output



import beignet
import torch


def test_legendre_polynomial_roots():
    torch.testing.assert_close(
        beignet.legendre_polynomial_roots(
            torch.tensor([1.0]),
        ),
        torch.tensor([]),
    )

    torch.testing.assert_close(
        beignet.legendre_polynomial_roots(torch.tensor([1.0, 2.0])),
        torch.tensor([-0.5]),
    )

    for index in range(2, 5):
        torch.testing.assert_close(
            beignet.trim_legendre_polynomial_coefficients(
                beignet.legendre_polynomial_roots(
                    beignet.legendre_polynomial_from_roots(
                        torch.linspace(-1, 1, index),
                    ),
                ),
                tol=0.000001,
            ),
            beignet.trim_legendre_polynomial_coefficients(
                torch.linspace(-1, 1, index),
                tol=0.000001,
            ),
        )



import beignet
import hypothesis.strategies
import torch
from scipy.spatial.transform import Rotation


@hypothesis.strategies.composite
def _strategy(function):
    size = function(
        hypothesis.strategies.integers(
            min_value=1,
            max_value=8,
        ),
    )

    rotation = Rotation.identity(size)

    return (
        {
            "size": size,
            "dtype": torch.float64,
        },
        torch.from_numpy(rotation.as_matrix()),
    )


@hypothesis.given(_strategy())
def test_rotation_matrix_identity(data):
    parameters, expected = data

    torch.testing.assert_close(
        beignet.rotation_matrix_identity(**parameters),
        expected,
    )



import torch
from torch import Tensor

from ._probabilists_hermite_polynomial_companion import (
    probabilists_hermite_polynomial_companion,
)


def probabilists_hermite_polynomial_roots(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    if input.shape[0] <= 1:
        return torch.tensor([], dtype=input.dtype)

    if input.shape[0] == 2:
        return torch.tensor([-input[0] / input[1]])

    output = probabilists_hermite_polynomial_companion(input)

    output = torch.flip(output, dims=[0])
    output = torch.flip(output, dims=[1])

    output = torch.linalg.eigvals(output)

    output, _ = torch.sort(output.real)

    return output



import math

import beignet
import torch


def test_gauss_physicists_hermite_polynomial_quadrature():
    x, w = beignet.gauss_physicists_hermite_polynomial_quadrature(100)

    v = beignet.physicists_hermite_polynomial_vandermonde(x, 99)
    vv = (v.T * w) @ v
    vd = 1 / torch.sqrt(vv.diagonal())
    vv = vd[:, None] * vv * vd
    torch.testing.assert_close(
        vv,
        torch.eye(100),
    )

    torch.testing.assert_close(
        w.sum(),
        torch.tensor(math.sqrt(math.pi)),
    )



import beignet
import torch


def test_fit_physicists_hermite_polynomial():
    def f(x):
        return x * (x - 1) * (x - 2)

    def g(x):
        return x**4 + x**2 + 1

    input = torch.linspace(0, 2, 50)

    other = f(input)

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=3,
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=torch.tensor([0, 1, 2, 3]),
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=4,
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=torch.tensor([0, 1, 2, 3, 4]),
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=torch.tensor([2, 3, 4, 1, 0]),
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            torch.stack([other, other]).T,
            degree=3,
        ),
        torch.stack(
            [
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
            ],
        ).T,
    )

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            torch.stack([other, other]).T,
            degree=torch.tensor([0, 1, 2, 3]),
        ),
        torch.stack(
            [
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
            ]
        ).T,
    )

    weight = torch.zeros_like(input)

    weight[1::2] = 1.0

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            other,
            degree=3,
            weight=weight,
        ),
        beignet.fit_physicists_hermite_polynomial(
            input,
            other,
            degree=torch.tensor([0, 1, 2, 3]),
        ),
    )

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            other,
            degree=torch.tensor([0, 1, 2, 3]),
            weight=weight,
        ),
        beignet.fit_physicists_hermite_polynomial(
            input,
            other,
            degree=torch.tensor([0, 1, 2, 3]),
        ),
    )

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            torch.stack([other, other]).T,
            degree=3,
            weight=weight,
        ),
        torch.stack(
            [
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
            ]
        ).T,
    )

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            torch.stack([other, other]).T,
            degree=torch.tensor([0, 1, 2, 3]),
            weight=weight,
        ),
        torch.stack(
            [
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
                (
                    beignet.fit_physicists_hermite_polynomial(
                        input,
                        other,
                        degree=torch.tensor([0, 1, 2, 3]),
                    )
                ),
            ]
        ).T,
    )

    # torch.testing.assert_close(
    #     beignet.hermfit(
    #         torch.tensor([1, 1j, -1, -1j]),
    #         torch.tensor([1, 1j, -1, -1j]),
    #         degree=1,
    #     ),
    #     torch.tensor([0.0j, 0.5j]),
    # )

    # torch.testing.assert_close(
    #     beignet.hermfit(
    #         torch.tensor([1, 1j, -1, -1j]),
    #         torch.tensor([1, 1j, -1, -1j]),
    #         degree=torch.tensor([0, 1]),
    #     ),
    #     torch.tensor([0, 0.5]),
    # )

    input = torch.linspace(-1, 1, 50)

    other = g(input)

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=4,
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial(
            input,
            beignet.fit_physicists_hermite_polynomial(
                input,
                other,
                degree=torch.tensor([0, 2, 4]),
            ),
        ),
        other,
    )

    torch.testing.assert_close(
        beignet.fit_physicists_hermite_polynomial(
            input,
            other,
            degree=4,
        ),
        beignet.fit_physicists_hermite_polynomial(
            input,
            other,
            degree=torch.tensor([0, 2, 4]),
        ),
    )



from torch import Tensor

from ._quaternion_to_rotation_vector import (
    quaternion_to_rotation_vector,
)
from ._rotation_matrix_to_quaternion import (
    rotation_matrix_to_quaternion,
)


def rotation_matrix_to_rotation_vector(
    input: Tensor,
    degrees: bool = False,
) -> Tensor:
    r"""
    Convert rotation matrices to rotation vectors.

    Parameters
    ----------
    input : Tensor, shape=(..., 3, 3)
        Rotation matrices.

    degrees : bool
        If `True`, rotation vector magnitudes are assumed to be in degrees.
        Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3)
        Rotation vectors.
    """
    return quaternion_to_rotation_vector(
        rotation_matrix_to_quaternion(
            input,
        ),
        degrees,
    )



import math

import torch
from torch import Tensor

from ._laguerre_polynomial_vandermonde import laguerre_polynomial_vandermonde


def fit_laguerre_polynomial(
    input: Tensor,
    other: Tensor,
    degree: Tensor | int,
    relative_condition: float | None = None,
    full: bool = False,
    weight: Tensor | None = None,
):
    input = torch.tensor(input)
    other = torch.tensor(other)
    degree = torch.tensor(degree)
    if degree.ndim > 1:
        raise TypeError
    # if deg.dtype.kind not in "iu":
    #     raise TypeError
    if math.prod(degree.shape) == 0:
        raise TypeError
    if degree.min() < 0:
        raise ValueError
    if input.ndim != 1:
        raise TypeError
    if input.size == 0:
        raise TypeError
    if other.ndim < 1 or other.ndim > 2:
        raise TypeError
    if len(input) != len(other):
        raise TypeError
    if degree.ndim == 0:
        lmax = int(degree)
        van = laguerre_polynomial_vandermonde(input, lmax)
    else:
        degree, _ = torch.sort(degree)
        lmax = int(degree[-1])
        van = laguerre_polynomial_vandermonde(input, lmax)[:, degree]
    # set up the least squares matrices in transposed form
    lhs = van.T
    rhs = other.T
    if weight is not None:
        if weight.ndim != 1:
            raise TypeError("expected 1D vector for w")

        if len(input) != len(weight):
            raise TypeError("expected x and w to have same length")

        # apply weights. Don't use inplace operations as they
        # can cause problems with NA.
        lhs = lhs * weight
        rhs = rhs * weight
    # set rcond
    if relative_condition is None:
        relative_condition = len(input) * torch.finfo(input.dtype).eps
    # Determine the norms of the design matrix columns.
    if torch.is_complex(lhs):
        scl = torch.sqrt((torch.square(lhs.real) + torch.square(lhs.imag)).sum(1))
    else:
        scl = torch.sqrt(torch.square(lhs).sum(1))
    scl = torch.where(scl == 0, 1, scl)
    # Solve the least squares problem.
    c, resids, rank, s = torch.linalg.lstsq(lhs.T / scl, rhs.T, relative_condition)
    c = (c.T / scl).T
    # Expand c to include non-fitted coefficients which are set to zero
    if degree.ndim > 0:
        if c.ndim == 2:
            cc = torch.zeros((lmax + 1, c.shape[1]), dtype=c.dtype)
        else:
            cc = torch.zeros(lmax + 1, dtype=c.dtype)

        cc[degree] = c

        c = cc
    if full:
        result = c, [resids, rank, s, relative_condition]
    else:
        result = c
    return result



import torch

from ._differentiate_laguerre_polynomial import differentiate_laguerre_polynomial
from ._evaluate_laguerre_polynomial import evaluate_laguerre_polynomial
from ._laguerre_polynomial_companion import laguerre_polynomial_companion


def gauss_laguerre_quadrature(degree):
    degree = int(degree)
    if degree <= 0:
        raise ValueError

    c = torch.zeros(degree + 1)
    c[-1] = 1.0

    m = laguerre_polynomial_companion(c)
    x = torch.linalg.eigvalsh(m)

    dy = evaluate_laguerre_polynomial(x, c)
    df = evaluate_laguerre_polynomial(x, differentiate_laguerre_polynomial(c))
    x = x - (dy / df)

    fm = evaluate_laguerre_polynomial(x, c[1:])
    fm = fm / torch.abs(fm).max()
    df = df / torch.abs(df).max()
    w = 1 / (fm * df)

    w = w / torch.sum(w)

    return x, w



import beignet
import torch


def test_legendre_polynomial_zero():
    torch.testing.assert_close(
        beignet.legendre_polynomial_zero,
        torch.tensor([0.0]),
        check_dtype=False,
    )



import beignet
import pytest
import torch


def test_legendre_polynomial_vandermonde():
    x = torch.arange(3)

    v = beignet.legendre_polynomial_vandermonde(
        x,
        degree=3,
    )

    assert v.shape == (3, 4)

    for index in range(4):
        torch.testing.assert_close(
            v[..., index],
            beignet.evaluate_legendre_polynomial(
                x,
                torch.tensor([0.0] * index + [1.0]),
            ),
        )

    x = torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])

    v = beignet.legendre_polynomial_vandermonde(
        x,
        degree=3,
    )

    assert v.shape == (3, 2, 4)

    for index in range(4):
        torch.testing.assert_close(
            v[..., index],
            beignet.evaluate_legendre_polynomial(
                x,
                torch.tensor([0.0] * index + [1.0]),
            ),
        )

    with pytest.raises(ValueError):
        beignet.legendre_polynomial_vandermonde(
            torch.tensor([1, 2, 3]),
            -1,
        )



from torch import Tensor

from ._quaternion_to_euler_angle import (
    quaternion_to_euler_angle,
)
from ._rotation_vector_to_quaternion import (
    rotation_vector_to_quaternion,
)


def rotation_vector_to_euler_angle(
    input: Tensor,
    axes: str,
    degrees: bool = False,
) -> Tensor:
    r"""
    Convert rotation vectors to Euler angles.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Rotation vectors.

    degrees : bool, optional
        If `True`, rotation vector magnitudes are assumed to be in degrees.
        Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3)
        Euler angles. The returned Euler angles are in the range:

            * First angle: :math:`(-180, 180]` degrees (inclusive)
            * Second angle:
                * :math:`[-90, 90]` degrees if all axes are different
                  (e.g., :math:`xyz`)
                * :math:`[0, 180]` degrees if first and third axes are the same
                  (e.g., :math:`zxz`)
            * Third angle: :math:`[-180, 180]` degrees (inclusive)
    """
    return quaternion_to_euler_angle(
        rotation_vector_to_quaternion(
            input,
            degrees,
        ),
        axes,
        degrees,
    )



import functools

import beignet
import torch


# def test_chebyshev_polynomial_power():
#     for j in range(5):
#         for k in range(5):
#             torch.testing.assert_close(
#                 beignet.trim_chebyshev_polynomial_coefficients(
#                     beignet.chebyshev_polynomial_power(
#                         torch.arange(0.0, j + 1),
#                         k,
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_chebyshev_polynomial_coefficients(
#                     functools.reduce(
#                         beignet.multiply_chebyshev_polynomial,
#                         [torch.arange(0.0, j + 1)] * k,
#                         torch.tensor([1.0]),
#                     ),
#                     tol=0.000001,
#                 ),
#             )



import torch
from torch import Tensor


def evaluate_legendre_polynomial(
    input: Tensor,
    coefficients: Tensor,
    tensor: bool = True,
) -> Tensor:
    coefficients = torch.atleast_1d(coefficients)

    if tensor:
        coefficients = torch.reshape(
            coefficients,
            coefficients.shape + (1,) * input.ndim,
        )

    match coefficients.shape[0]:
        case 1:
            a = coefficients[0]
            b = 0.0
        case 2:
            a = coefficients[0]
            b = coefficients[1]
        case _:
            size = coefficients.shape[0]

            a = coefficients[-2] * torch.ones_like(input)
            b = coefficients[-1] * torch.ones_like(input)

            for index in range(3, coefficients.shape[0] + 1):
                previous = a

                size = size - 1

                a = coefficients[-index] - (b * (size - 1.0)) / size

                b = previous + (b * input * (2.0 * size - 1.0)) / size

    return a + b * input



import torch
from torch import Tensor


def quaternion_to_rotation_matrix(input: Tensor) -> Tensor:
    r"""
    Convert rotation quaternions to rotation matrices.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    Returns
    -------
    output : Tensor, shape=(..., 3, 3)
        Rotation matrices.
    """
    output = torch.empty(
        [input.shape[0], 3, 3],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(input.shape[0]):
        a = input[j, 0]
        b = input[j, 1]
        c = input[j, 2]
        d = input[j, 3]

        output[j, 0, 0] = +(a**2.0) - b**2.0 - c**2.0 + d**2.0
        output[j, 1, 1] = -(a**2.0) + b**2.0 - c**2.0 + d**2.0
        output[j, 2, 2] = -(a**2.0) - b**2.0 + c**2.0 + d**2.0

        output[j, 0, 1] = 2.0 * (a * b) - 2.0 * (c * d)
        output[j, 0, 2] = 2.0 * (a * c) + 2.0 * (b * d)
        output[j, 1, 0] = 2.0 * (a * b) + 2.0 * (c * d)
        output[j, 1, 2] = 2.0 * (b * c) - 2.0 * (a * d)
        output[j, 2, 0] = 2.0 * (a * c) - 2.0 * (b * d)
        output[j, 2, 1] = 2.0 * (b * c) + 2.0 * (a * d)

    return output



from typing import Tuple

import torch
from torch import Tensor

from ._multiply_probabilists_hermite_polynomial import (
    multiply_probabilists_hermite_polynomial,
)


def divide_probabilists_hermite_polynomial(
    input: Tensor,
    other: Tensor,
) -> Tuple[Tensor, Tensor]:
    r"""
    Returns the quotient and remainder of two polynomials.

    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    other : Tensor
        Polynomial coefficients.

    Returns
    -------
    output : Tuple[Tensor, Tensor]
        Polynomial coefficients of the quotient and remainder.
    """
    input = torch.atleast_1d(input)
    other = torch.atleast_1d(other)

    dtype = torch.promote_types(input.dtype, other.dtype)

    input = input.to(dtype)
    other = other.to(dtype)

    m = input.shape[0]
    n = other.shape[0]

    if m < n:
        return torch.zeros_like(input[:1]), input

    if n == 1:
        return input / other[-1], torch.zeros_like(input[:1])

    def f(x: Tensor) -> Tensor:
        indicies = torch.flip(x, [0])

        indicies = torch.nonzero(indicies, as_tuple=False)

        if indicies.shape[0] > 1:
            indicies = indicies[:1]

        if indicies.shape[0] < 1:
            indicies = torch.concatenate(
                [
                    indicies,
                    torch.full(
                        [
                            1 - indicies.shape[0],
                            indicies.shape[1],
                        ],
                        0,
                    ),
                ],
                0,
            )

        return x.shape[0] - 1 - indicies[0][0]

    quotient = torch.zeros(m - n + 1, dtype=input.dtype)

    ridx = input.shape[0] - 1

    size = m - f(other) - 1

    y = torch.zeros(m + n + 1, dtype=input.dtype)

    y[size] = 1.0

    x = quotient, input, y, ridx

    for index in range(0, size):
        quotient, remainder, y2, ridx1 = x

        j = size - index

        p = multiply_probabilists_hermite_polynomial(y2, other)

        pidx = f(p)

        t = remainder[ridx1] / p[pidx]

        remainder_modified = remainder.clone()
        remainder_modified[ridx1] = 0.0

        a = remainder_modified

        p_modified = p.clone()
        p_modified[pidx] = 0.0

        b = t * p_modified

        a = torch.atleast_1d(a)
        b = torch.atleast_1d(b)

        dtype = torch.promote_types(a.dtype, b.dtype)

        a = a.to(dtype)
        b = b.to(dtype)

        if a.shape[0] > b.shape[0]:
            output = -b

            output = torch.concatenate(
                [
                    output,
                    torch.zeros(
                        a.shape[0] - b.shape[0],
                        dtype=b.dtype,
                    ),
                ],
            )
            output = a + output
        else:
            output = -b

            output = torch.concatenate(
                [
                    output[: a.shape[0]] + a,
                    output[a.shape[0] :],
                ],
            )

        remainder = output

        remainder = remainder[: remainder.shape[0]]

        quotient[j] = t

        ridx1 = ridx1 - 1

        y2 = torch.roll(y2, -1)

        x = quotient, remainder, y2, ridx1

    quotient, remainder, _, _ = x

    return quotient, remainder



import torch
from torch import Tensor


def legendre_polynomial_vandermonde(x: Tensor, degree: Tensor) -> Tensor:
    if degree < 0:
        raise ValueError

    x = torch.tensor(x)
    x = torch.atleast_1d(x)

    dims = (degree + 1,) + x.shape

    dtype = torch.promote_types(x.dtype, torch.tensor(0.0).dtype)

    x = x.to(dtype)

    v = torch.empty(dims, dtype=dtype)

    v[0] = torch.ones_like(x)

    if degree > 0:
        v[1] = x

        for index in range(2, degree + 1):
            v[index] = (
                v[index - 1] * x * (2 * index - 1) - v[index - 2] * (index - 1)
            ) / index

    return torch.moveaxis(v, 0, -1)



import torch
from torch import Tensor

from ._multiply_physicists_hermite_polynomial import (
    multiply_physicists_hermite_polynomial,
)


def physicists_hermite_polynomial_power(
    input: Tensor,
    exponent: float | Tensor,
    maximum_exponent: float | Tensor = 16.0,
) -> Tensor:
    input = torch.atleast_1d(input)
    _exponent = int(exponent)
    if _exponent != exponent or _exponent < 0:
        raise ValueError
    if maximum_exponent is not None and _exponent > maximum_exponent:
        raise ValueError
    match _exponent:
        case 0:
            output = torch.tensor([1], dtype=input.dtype)
        case 1:
            output = input
        case _:
            output = torch.zeros(input.shape[0] * exponent, dtype=input.dtype)

            input = torch.atleast_1d(input)
            output = torch.atleast_1d(output)

            dtype = torch.promote_types(input.dtype, output.dtype)

            input = input.to(dtype)
            output = output.to(dtype)

            if output.shape[0] > input.shape[0]:
                input = torch.concatenate(
                    [
                        input,
                        torch.zeros(
                            output.shape[0] - input.shape[0],
                            dtype=input.dtype,
                        ),
                    ],
                )

                output = output + input
            else:
                output = torch.concatenate(
                    [
                        output,
                        torch.zeros(
                            input.shape[0] - output.shape[0],
                            dtype=output.dtype,
                        ),
                    ]
                )

                output = input + output

            for _ in range(2, _exponent + 1):
                output = multiply_physicists_hermite_polynomial(
                    output, input, mode="same"
                )
    return output



import beignet
import torch


def test_evaluate_laguerre_polynomial_cartesian_3d():
    c1d = torch.tensor([9.0, -14.0, 6.0])
    c3d = torch.einsum("i,j,k->ijk", c1d, c1d, c1d)

    x = torch.rand(3, 5) * 2 - 1
    y = beignet.evaluate_polynomial(x, torch.tensor([1.0, 2.0, 3.0]))

    a, b, x3 = x
    y1, y2, y3 = y

    target = torch.einsum("i,j,k->ijk", y1, y2, y3)
    torch.testing.assert_close(
        beignet.evaluate_laguerre_polynomial_cartesian_3d(a, b, x3, c3d), target
    )

    z = torch.ones([2, 3])
    assert (
        beignet.evaluate_laguerre_polynomial_cartesian_3d(z, z, z, c3d).shape
        == (2, 3) * 3
    )



import numpy
import torch
import torch._numpy._funcs_impl
from torch import Tensor

from ._evaluate_polynomial import evaluate_polynomial


def integrate_polynomial(
    input: Tensor,
    order: int = 1,
    k=None,
    lower_bound: float = 0,
    scale: float = 1,
    dim: int = 0,
) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor
        Polynomial coefficients.

    order : int, default=1

    k : int, optional

    lower_bound : float, default=0

    scale : float, default=1

    dim : int, default=0

    Returns
    -------
    output : Tensor
        Polynomial coefficients of the integral.
    """
    if k is None:
        k = []

    input = torch.atleast_1d(input)

    lower_bound, scale = map(torch.tensor, (lower_bound, scale))

    if not numpy.iterable(k):
        k = [k]

    if order < 0:
        raise ValueError

    if len(k) > order:
        raise ValueError

    if lower_bound.ndim != 0:
        raise ValueError

    if scale.ndim != 0:
        raise ValueError

    if order == 0:
        return input

    k = torch.tensor(list(k) + [0] * (order - len(k)))
    k = torch.atleast_1d(k)

    n = input.shape[dim]

    padding = torch.tensor([0, order])
    input = torch.moveaxis(input, dim, 0)
    if padding[0] < 0:
        input = input[torch.abs(padding[0]) :]

        padding = (0, padding[1])
    if padding[1] < 0:
        input = input[: -torch.abs(padding[1])]

        padding = (padding[0], 0)
    npad = torch.tensor([(0, 0)] * input.ndim)
    npad[0] = padding
    output = torch._numpy._funcs_impl.pad(
        input,
        pad_width=npad,
        mode="constant",
        constant_values=0,
    )
    input = torch.moveaxis(output, 0, dim)

    input = torch.moveaxis(input, dim, 0)

    d = torch.arange(n + order) + 1

    for i in range(0, order):
        input = input * scale

        input = (input.T / d).T

        input = torch.roll(input, 1, dims=[0])

        input[0] = 0.0

        input[0] += k[i] - evaluate_polynomial(lower_bound, input)

    return torch.moveaxis(input, 0, dim)



from torch import Tensor

from ._compose_quaternion import compose_quaternion
from ._euler_angle_to_quaternion import euler_angle_to_quaternion
from ._quaternion_to_euler_angle import quaternion_to_euler_angle


def compose_euler_angle(
    input: Tensor,
    other: Tensor,
    axes: str,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Compose rotation quaternions.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Euler angles.

    other : Tensor, shape=(..., 3)
        Euler angles.

    axes : str
        Axes. One to three characters belonging to the set :math:`\{X, Y, Z\}`
        for intrinsic rotations, or :math:`\{x, y, z\}` for extrinsic
        rotations. Extrinsic and intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees. Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3)
        Composed Euler angles.
    """
    return quaternion_to_euler_angle(
        compose_quaternion(
            euler_angle_to_quaternion(
                input,
                axes,
                degrees,
            ),
            euler_angle_to_quaternion(
                other,
                axes,
                degrees,
            ),
        ),
        axes,
        degrees,
    )



from torch import Tensor

from ._evaluate_laguerre_polynomial import evaluate_laguerre_polynomial


def evaluate_laguerre_polynomial_cartesian_3d(
    x: Tensor,
    y: Tensor,
    z: Tensor,
    c: Tensor,
) -> Tensor:
    for arg in [x, y, z]:
        c = evaluate_laguerre_polynomial(arg, c)
    return c



from pathlib import Path
from typing import Callable, Tuple, Union

import torch
from pandas import DataFrame
from torch import Tensor

from beignet.transforms import Transform

from ._atom3d_dataset import ATOM3DDataset


class ATOM3DMSPDataset(ATOM3DDataset):
    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = False,
        transform: Union[Callable, Transform, None] = None,
        target_transform: Union[Callable, Transform, None] = None,
    ):
        super().__init__(
            root,
            "raw/MSP/data",
            "https://zenodo.org/record/4962515/files/MSP-raw.tar.gz",
            "MSP",
            checksum="77aeb79cfc80bd51cdfb2aa321bf6128",
            download=download,
        )

        self._transform_fn = transform

        self._target_transform_fn = target_transform

    def __getitem__(
        self,
        index: int,
    ) -> Tuple[Tuple[DataFrame, DataFrame], Tensor]:
        item = super().__getitem__(index)

        structure = DataFrame(**item["original_atoms"])

        mutant = DataFrame(**item["mutated_atoms"])

        if self._transform_fn is not None:
            structure, mutant = self._transform_fn(structure, mutant)

        target = torch.tensor(int(item["label"]))

        if self._target_transform_fn is not None:
            target = self._target_transform_fn(target)

        return (structure, mutant), target



import math

import torch
from torch import Tensor


def _voigt_v(x, y, n: int = 11):
    if not (((x >= 0.0) & (y >= 0.0)) | torch.isnan(x) | torch.isnan(y)).all():
        raise ValueError("_voigt_v only defined for x >= 0 and y >= 0")

    h = math.sqrt(math.pi / (n + 1))

    phi = (x / h) - torch.floor(x / h)

    k = torch.arange(n + 1, dtype=x.dtype, device=x.device)
    t = (k + 0.5) * h
    tau = k[1:] * h

    # equation 12
    w_m = (2 * h * y / torch.pi) * (
        torch.exp(-t.pow(2))
        * (t.pow(2) + x[..., None].pow(2) + y[..., None].pow(2))
        / (
            ((t - x[..., None]).pow(2) + y[..., None].pow(2))
            * ((t + x[..., None]).pow(2) + y[..., None].pow(2))
        )
    ).sum(dim=-1)

    # equation 13
    expy = torch.exp(-2 * torch.pi * y / h)
    w_mm_1 = (
        2
        * torch.exp(-x.pow(2) + y.pow(2))
        * (torch.cos(2 * x * y) * expy + torch.cos(2 * torch.pi * x / h - 2 * x * y))
    ) / (expy + 1 / expy + 2 * torch.cos(2 * torch.pi * x / h))

    w_mm = w_mm_1 + w_m

    w_mt_1 = (
        2
        * torch.exp(-x.pow(2) + y.pow(2))
        * (torch.cos(2 * x * y) * expy - torch.cos(2 * torch.pi * x / h - 2 * x * y))
    ) / (expy + 1 / expy - 2 * torch.cos(2 * torch.pi * x / h))

    w_mt_2 = (h * y) / (torch.pi * (x.pow(2) + y.pow(2)))

    w_mt_3 = (2 * h * y / torch.pi) * (
        torch.exp(-tau.pow(2))
        * (tau.pow(2) + x[..., None].pow(2) + y[..., None].pow(2))
        / (
            ((tau - x[..., None]).pow(2) + y[..., None].pow(2))
            * ((tau + x[..., None]).pow(2) + y[..., None].pow(2))
        )
    ).sum(dim=-1)

    # equation 14
    w_mt = w_mt_1 + w_mt_2 + w_mt_3

    return torch.where(
        y >= torch.maximum(x, torch.tensor(torch.pi / h)),
        w_m,
        torch.where((y < x) & (1 / 4 <= phi) & (phi <= 3 / 4), w_mt, w_mm),
    )


def _voigt_l(x, y, n: int = 11):
    if not (((x >= 0.0) & (y >= 0.0)) | torch.isnan(x) | torch.isnan(y)).all():
        raise ValueError("_voigt_l only defined for x >= 0 and y >= 0")

    h = math.sqrt(math.pi / (n + 1))

    phi = (x / h) - torch.floor(x / h)

    k = torch.arange(n + 1, dtype=x.dtype, device=x.device)
    t = (k + 0.5) * h
    tau = k[1:] * h

    w_m = (2 * h * x / torch.pi) * (
        torch.exp(-t.pow(2))
        * (-t.pow(2) + x[..., None].pow(2) + y[..., None].pow(2))
        / (
            ((t - x[..., None]).pow(2) + y[..., None].pow(2))
            * ((t + x[..., None]).pow(2) + y[..., None].pow(2))
        )
    ).sum(dim=-1)

    # equation 13
    expy = torch.exp(-2 * torch.pi * y / h)
    w_mm_1 = (
        -2
        * torch.exp(-x.pow(2) + y.pow(2))
        * (torch.sin(2 * x * y) * expy - torch.sin(2 * torch.pi * x / h - 2 * x * y))
    ) / (expy + 1 / expy + 2 * torch.cos(2 * torch.pi * x / h))

    w_mm = w_mm_1 + w_m

    w_mt_1 = (
        -2
        * torch.exp(-x.pow(2) + y.pow(2))
        * (torch.sin(2 * x * y) * expy + torch.sin(2 * torch.pi * x / h - 2 * x * y))
    ) / (expy + 1 / expy - 2 * torch.cos(2 * torch.pi * x / h))

    w_mt_2 = (h * x) / (torch.pi * (x.pow(2) + y.pow(2)))

    w_mt_3 = (2 * h * x / torch.pi) * (
        torch.exp(-tau.pow(2))
        * (-tau.pow(2) + x[..., None].pow(2) + y[..., None].pow(2))
        / (
            ((tau - x[..., None]).pow(2) + y[..., None].pow(2))
            * ((tau + x[..., None]).pow(2) + y[..., None].pow(2))
        )
    ).sum(dim=-1)

    # equation 14
    w_mt = w_mt_1 + w_mt_2 + w_mt_3

    return torch.where(
        y >= torch.maximum(x, torch.tensor(torch.pi / h)),
        w_m,
        torch.where((y < x) & (1 / 4 <= phi) & (phi <= 3 / 4), w_mt, w_mm),
    )


def faddeeva_w(input: Tensor, *, out: Tensor | None = None) -> Tensor:
    r"""
    Faddeeva function.

    Parameters
    ----------
    input : Tensor
        Input tensor.

    out : Tensor, optional
        Output tensor.

    Returns
    -------
    Tensor
    """

    if not torch.is_complex(input):
        input = torch.complex(input, torch.zeros_like(input))

    # use symmetries to map to upper right quadrant of complex plane
    imag_negative = input.imag < 0.0
    input = torch.where(input.imag < 0.0, -input, input)
    real_negative = input.real < 0.0
    input = torch.where(input.real < 0.0, -input.conj(), input)

    x = input.real
    y = input.imag

    if not (((x >= 0.0) & (y >= 0.0)) | torch.isnan(x) | torch.isnan(y)).all():
        raise ValueError("failed to map input to x >= 0, y >= 0")

    output = _voigt_v(x, y, n=11) + 1j * _voigt_l(x, y, n=11)

    # compute real and imaginary parts separately to so we handle infs
    # without unnecessary nans
    expz2 = torch.complex(
        2 * torch.exp(-x.pow(2) + y.pow(2)) * torch.cos(-2 * x * y),
        2 * torch.exp(-x.pow(2) + y.pow(2)) * torch.sin(-2 * x * y),
    )
    output = torch.where(imag_negative, expz2 - output, output)
    output = torch.where(real_negative, output.conj(), output, out=out)

    if out is not None:
        out.copy_(output)
        return out
    else:
        return output



# import torch
# import torch.func
# from beignet import apply_transform
# from torch import Tensor
#
#
# def test_apply_transform():
#     input = torch.randn([32, 3])
#
#     transform = torch.randn([3, 3])
#
#     def f(r: Tensor) -> Tensor:
#         return torch.sum(r**2)
#
#     def g(r: Tensor, t: Tensor) -> Tensor:
#         return torch.sum(apply_transform(r, t) ** 2)
#
#     torch.testing.assert_close(
#         torch.func.grad(f)(apply_transform(input, transform)),
#         torch.func.grad(g, 0)(input, transform),
#     )



import beignet
import torch


# def test_polynomial_roots():
#     torch.testing.assert_close(
#         beignet.polynomial_roots(torch.tensor([1.0])),
#         torch.tensor([]),
#     )
#
#     torch.testing.assert_close(
#         beignet.polynomial_roots(torch.tensor([1.0, 2.0])),
#         torch.tensor([-0.5]),
#     )
#
#     for index in range(2, 5):
#         input = torch.linspace(-1, 1, index)
#
#         torch.testing.assert_close(
#             beignet.trim_polynomial_coefficients(
#                 beignet.polynomial_roots(
#                     beignet.polynomial_from_roots(
#                         input,
#                     ),
#                 ),
#                 tol=0.000001,
#             ),
#             beignet.trim_polynomial_coefficients(
#                 input,
#                 tol=0.000001,
#             ),
#         )



import re

import torch
from torch import Tensor


def euler_angle_to_quaternion(
    input: Tensor,
    axes: str,
    degrees: bool = False,
    canonical: bool | None = False,
) -> Tensor:
    r"""
    Convert Euler angles to rotation quaternions.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Euler angles.

    axes : str
        Axes. One to three characters belonging to the set :math:`\{X, Y, Z\}`
        for intrinsic rotations, or :math:`\{x, y, z\}` for extrinsic
        rotations. Extrinsic and intrinsic rotations cannot be mixed.

    degrees : bool, optional
        If `True`, Euler angles are assumed to be in degrees. Default, `False`.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    Returns
    -------
    output : Tensor, shape=(..., 4)
        Rotation quaternions.
    """
    intrinsic = re.match(r"^[XYZ]{1,3}$", axes) is not None

    if degrees:
        input = torch.deg2rad(input)

    if len(axes) == 1:
        if input.ndim == 0:
            input = input.reshape([1, 1])
        elif input.ndim == 1:
            input = input[:, None]
        elif input.ndim == 2 and input.shape[-1] != 1:
            raise ValueError
        elif input.ndim > 2:
            raise ValueError
    else:
        if input.ndim not in [1, 2] or input.shape[-1] != len(axes):
            raise ValueError

        if input.ndim == 1:
            input = input[None, :]

    if input.ndim != 2 or input.shape[-1] != len(axes):
        raise ValueError

    output = torch.zeros(
        [input.shape[0], 4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    match axes.lower()[0]:
        case "x":
            k = 0
        case "y":
            k = 1
        case "z":
            k = 2
        case _:
            raise ValueError

    for j in range(input[:, 0].shape[0]):
        output[j, 3] = torch.cos(input[:, 0][j] / 2)
        output[j, k] = torch.sin(input[:, 0][j] / 2)

    z = output

    c = torch.empty(
        [3],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(1, len(axes.lower())):
        y = torch.zeros(
            [input.shape[0], 4],
            dtype=input.dtype,
            layout=input.layout,
            device=input.device,
        )

        r = torch.empty(
            [max(y.shape[0], z.shape[0]), 4],
            dtype=input.dtype,
            layout=input.layout,
            device=input.device,
        )

        match axes.lower()[j]:
            case "x":
                p = 0
            case "y":
                p = 1
            case "z":
                p = 2
            case _:
                raise ValueError

        for k in range(input[:, j].shape[0]):
            y[k, 3] = torch.cos(input[:, j][k] / 2)
            y[k, p] = torch.sin(input[:, j][k] / 2)

        if intrinsic:
            for k in range(max(y.shape[0], z.shape[0])):
                c[0] = z[k, 1] * y[k, 2] - z[k, 2] * y[k, 1]
                c[1] = z[k, 2] * y[k, 0] - z[k, 0] * y[k, 2]
                c[2] = z[k, 0] * y[k, 1] - z[k, 1] * y[k, 0]

                t = z[k, 0]
                u = z[k, 1]
                v = z[k, 2]
                w = z[k, 3]

                r[k, 0] = w * y[k, 0] + y[k, 3] * t + c[0]
                r[k, 1] = w * y[k, 1] + y[k, 3] * u + c[1]
                r[k, 2] = w * y[k, 2] + y[k, 3] * v + c[2]
                r[k, 3] = w * y[k, 3] - t * y[k, 0] - u * y[k, 1] - v * y[k, 2]

            z = r
        else:
            for k in range(max(y.shape[0], z.shape[0])):
                c[0] = y[k, 1] * z[k, 2] - y[k, 2] * z[k, 1]
                c[1] = y[k, 2] * z[k, 0] - y[k, 0] * z[k, 2]
                c[2] = y[k, 0] * z[k, 1] - y[k, 1] * z[k, 0]

                t = z[k, 0]
                u = z[k, 1]
                v = z[k, 2]
                w = z[k, 3]

                r[k, 0] = y[k, 3] * t + w * y[k, 0] + c[0]
                r[k, 1] = y[k, 3] * u + w * y[k, 1] + c[1]
                r[k, 2] = y[k, 3] * v + w * y[k, 2] + c[2]
                r[k, 3] = y[k, 3] * w - y[k, 0] * t - y[k, 1] * u - y[k, 2] * v

            z = r

    if canonical:
        for j in range(z.shape[0]):
            a = z[j, 0]
            b = z[j, 1]
            c = z[j, 2]
            d = z[j, 3]

            if d == 0 and (a == 0 & (b == 0 & c < 0 | b < 0) | a < 0) | d < 0:
                z[j] = -z[j]

    return z



import beignet
import pytest
import torch


def test_laguerre_polynomial_companion():
    with pytest.raises(ValueError):
        beignet.laguerre_polynomial_companion(
            torch.tensor([]),
        )

    with pytest.raises(ValueError):
        beignet.laguerre_polynomial_companion(
            torch.tensor([1.0]),
        )

    for index in range(1, 5):
        output = beignet.laguerre_polynomial_companion(
            torch.tensor([0.0] * index + [1.0]),
        )

        assert output.shape == (index, index)

    output = beignet.laguerre_polynomial_companion(
        torch.tensor([1.0, 2.0]),
    )

    assert output[0, 0] == 1.5



import torch
from torch import Tensor


def quaternion_mean(
    input: Tensor,
    weight: Tensor | None = None,
) -> Tensor:
    r"""
    Mean rotation quaternions.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    weight : Tensor, shape=(..., 4), optional
        Relative importance of rotation quaternions.

    Returns
    -------
    output : Tensor, shape=(..., 4)
        Rotation quaternions mean.
    """
    if weight is None:
        weight = torch.ones(input.shape[0])

    _, output = torch.linalg.eigh((input.T * weight) @ input)

    output = output[:, -1]

    output = torch.unsqueeze(output, dim=0)

    return output



import torch
from torch import Tensor


def polynomial_vandermonde(input: Tensor, degree: Tensor) -> Tensor:
    r"""
    Parameters
    ----------
    input : Tensor

    degree : Tensor

    Returns
    -------
    output : Tensor
    """
    if degree < 0:
        raise ValueError

    degree = int(degree)

    input = torch.atleast_1d(input)

    output = torch.empty([degree + 1, *input.shape], dtype=input.dtype)

    output[0] = torch.ones_like(input)

    for i in range(1, degree + 1):
        output[i] = output[i - 1] * input

    output = torch.moveaxis(output, 0, -1)

    return output



# import beignet
# import torch
#
#
# def test_laguerre_polynomial_weight():
#     torch.testing.assert_close(
#         beignet.laguerre_polynomial_weight(
#             torch.linspace(0, 10, 11),
#         ),
#         torch.exp(-torch.linspace(0, 10, 11)),
#     )



import torch
from beignet import farthest_first_traversal


def str_hamming_dist(s1, s2):
    return sum(c1 != c2 for c1, c2 in zip(s1, s2, strict=False))


def ell_2_dist(v1, v2):
    return torch.norm(v1 - v2, p=2)


def test_farthest_first_traversal_str():
    library = [
        "AAAA",
        "GGGG",
        "CCCC",
        "TTTT",
    ]
    ranking_scores = torch.tensor([3, 2, 1, 4])
    n = 2
    selected = farthest_first_traversal(
        library, str_hamming_dist, ranking_scores, n, descending=True
    )

    assert torch.all(selected == torch.tensor([3, 0]))

    selected = farthest_first_traversal(
        library, str_hamming_dist, ranking_scores, n, descending=False
    )
    assert torch.all(selected == torch.tensor([2, 1]))

    ranking_scores = None
    selected = farthest_first_traversal(library, str_hamming_dist, ranking_scores, n)
    assert torch.all(selected == torch.tensor([0, 1]))

    # harder example with wider spread of distances
    library = [
        "AAAA",
        "GGGG",
        "CCCC",
        "TTTT",
        "ACGT",
        "TGCA",
        "ACGT",
        "TGCA",
    ]

    ranking_scores = torch.tensor(list(range(8)))
    n = 3
    selected = farthest_first_traversal(
        library, str_hamming_dist, ranking_scores, n, descending=True
    )
    assert torch.all(selected == torch.tensor([7, 6, 3]))

    selected = farthest_first_traversal(
        library, str_hamming_dist, ranking_scores, n, descending=False
    )
    assert torch.all(selected == torch.tensor([0, 1, 2]))


def test_farthest_first_traversal_tensor():
    library = torch.tensor(
        [
            [0.0, 0.0],
            [0.0, 4.0],
            [4.0, 0.0],
            [2.0, 2.0],
        ]
    )
    ranking_scores = torch.tensor([3, 2, 1, 4])
    n = 2
    selected = farthest_first_traversal(
        library, ell_2_dist, ranking_scores, n, descending=True
    )

    assert torch.all(selected == torch.tensor([3, 0]))

    selected = farthest_first_traversal(
        library, ell_2_dist, ranking_scores, n, descending=False
    )
    assert torch.all(selected == torch.tensor([2, 1]))

    ranking_scores = None
    selected = farthest_first_traversal(library, ell_2_dist, ranking_scores, n)
    assert torch.all(selected == torch.tensor([0, 1]))



import numpy
import torch
from torch import Tensor

from ._evaluate_chebyshev_polynomial import evaluate_chebyshev_polynomial


def integrate_chebyshev_polynomial(
    input: Tensor,
    order=1,
    k=None,
    lower_bound=0,
    scale=1,
    axis=0,
) -> Tensor:
    if k is None:
        k = []

    input = torch.atleast_1d(input)

    lower_bound = torch.tensor(lower_bound)

    scale = torch.tensor(scale)

    if not numpy.iterable(k):
        k = [k]

    if len(k) > order:
        raise ValueError

    if lower_bound.ndim != 0:
        raise ValueError

    if scale.ndim != 0:
        raise ValueError

    if order == 0:
        return input

    input = torch.moveaxis(input, axis, 0)

    k = torch.tensor([*k] + [0.0] * (order - len(k)))

    k = torch.atleast_1d(k)

    for i in range(order):
        n = input.shape[0]

        input = input * scale

        tmp = torch.empty([n + 1, *input.shape[1:]])

        tmp[0] = input[0] * 0
        tmp[1] = input[0]

        if n > 1:
            tmp[2] = input[1] / 4

        if n < 2:
            j = torch.tensor([], dtype=torch.int32)
        else:
            j = torch.arange(2, n)

        tmp[j + 1] = (input[j].T / (2 * (j + 1))).T
        tmp[j - 1] = tmp[j - 1] + -(input[j] / (2 * (j - 1)))

        tmp[0] = tmp[0] + (k[i] - evaluate_chebyshev_polynomial(lower_bound, tmp))

        input = tmp

    input = torch.moveaxis(input, 0, axis)

    return input



import beignet
import torch


def test_linear_physicists_hermite_polynomial():
    torch.testing.assert_close(
        beignet.linear_physicists_hermite_polynomial(3, 4),
        torch.tensor([3.0, 2.0]),
    )



import beignet
import pytest
import torch


def test_chebyshev_zeros():
    with pytest.raises(ValueError):
        beignet.chebyshev_zeros(0)

    torch.testing.assert_close(
        beignet.chebyshev_zeros(1),
        torch.tensor([0.0]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_zeros(2),
        torch.tensor([-0.70710678118654746, 0.70710678118654746]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_zeros(3),
        torch.tensor([-0.86602540378443871, 0, 0.86602540378443871]),
    )

    torch.testing.assert_close(
        beignet.chebyshev_zeros(4),
        torch.tensor([-0.9238795325, -0.3826834323, 0.3826834323, 0.9238795325]),
    )



import torch
from torch import Tensor

from ._add_polynomial import add_polynomial
from ._multiply_polynomial_by_x import multiply_polynomial_by_x
from ._subtract_polynomial import subtract_polynomial


def chebyshev_polynomial_to_polynomial(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    n = input.shape[0]

    if n < 3:
        return input

    c0 = torch.zeros_like(input)
    c0[0] = input[-2]

    c1 = torch.zeros_like(input)
    c1[0] = input[-1]

    for index in range(0, n - 2):
        i1 = n - 1 - index

        tmp = c0

        c0 = subtract_polynomial(input[i1 - 2], c1)

        c1 = add_polynomial(tmp, multiply_polynomial_by_x(c1, "same") * 2)

    output = multiply_polynomial_by_x(c1, "same")

    output = add_polynomial(c0, output)

    return output



import beignet
import torch


# def test_multiply_polynomial():
#     for j in range(5):
#         for k in range(5):
#             target = torch.zeros(j + k + 1)
#
#             target[j + k] = target[j + k] + 1
#
#             torch.testing.assert_close(
#                 beignet.trim_polynomial_coefficients(
#                     beignet.multiply_polynomial(
#                         torch.tensor([0.0] * j + [1.0]),
#                         torch.tensor([0.0] * k + [1.0]),
#                     ),
#                     tol=0.000001,
#                 ),
#                 beignet.trim_polynomial_coefficients(
#                     target,
#                     tol=0.000001,
#                 ),
#             )



import beignet
import torch


def test_multiply_physicists_hermite_polynomial():
    for i in range(5):
        input = torch.linspace(-3, 3, 100)

        val1 = beignet.evaluate_physicists_hermite_polynomial(
            input,
            torch.tensor([0.0] * i + [1.0]),
        )

        for j in range(5):
            val2 = beignet.evaluate_physicists_hermite_polynomial(
                input,
                torch.tensor([0.0] * j + [1.0]),
            )

            torch.testing.assert_close(
                beignet.evaluate_physicists_hermite_polynomial(
                    input,
                    beignet.multiply_physicists_hermite_polynomial(
                        torch.tensor([0.0] * i + [1.0]),
                        torch.tensor([0.0] * j + [1.0]),
                    ),
                ),
                val1 * val2,
            )



import math

import torch

from ._probabilists_hermite_polynomial_companion import (
    probabilists_hermite_polynomial_companion,
)


def gauss_probabilists_hermite_polynomial_quadrature(degree):
    degree = int(degree)
    if degree <= 0:
        raise ValueError

    c = torch.zeros(degree + 1)
    c[-1] = 1.0
    m = probabilists_hermite_polynomial_companion(c)
    x = torch.linalg.eigvalsh(m)

    if degree == 0:
        dy = torch.full(x.shape, 1.0 / math.sqrt(math.sqrt(2.0 * math.pi)))
    else:
        a1 = torch.zeros_like(x)
        b1 = torch.ones_like(x) / math.sqrt(math.sqrt(2.0 * math.pi))

        size = torch.tensor(degree)

        for _ in range(0, degree - 1):
            previous = a1

            a1 = -b1 * torch.sqrt((size - 1.0) / size)

            b1 = previous + b1 * x * torch.sqrt(1.0 / size)

            size = size - 1.0

        dy = a1 + b1 * x

    n = degree - 1

    if n == 0:
        df = torch.full(x.shape, 1 / math.sqrt(math.sqrt(math.pi)))
    else:
        a = torch.zeros_like(x)

        b = torch.ones_like(x) / math.sqrt(math.sqrt(math.pi))

        size = torch.tensor(n)

        for _ in range(0, n - 1):
            previous = a

            a = -b * torch.sqrt((size - 1.0) / size)

            b = previous + b * x * torch.sqrt(2.0 / size)

            size = size - 1.0

        df = a + b * x * math.sqrt(2.0)

    df = df * math.sqrt(degree)

    x = x - (dy / df)

    n = degree - 1

    if n == 0:
        fm = torch.full(x.shape, 1.0 / math.sqrt(math.sqrt(2.0 * math.pi)))
    else:
        a = torch.zeros_like(x)
        b = torch.ones_like(x) / math.sqrt(math.sqrt(2.0 * math.pi))

        size = torch.tensor(n)

        for _ in range(0, n - 1):
            previous = a

            a = -b * torch.sqrt((size - 1.0) / size)

            b = previous + b * x * torch.sqrt(1.0 / size)

            size = size - 1.0

        fm = a + b * x

    fm = fm / torch.abs(fm).max()

    w = 1 / (fm * fm)

    a = torch.flip(w, dims=[0])
    b = torch.flip(x, dims=[0])

    w = (w + a) / 2
    x = (x - b) / 2

    w = w * (math.sqrt(2 * math.pi) / torch.sum(w))

    return x, w



import beignet
import torch


def test_subtract_chebyshev_polynomial():
    for j in range(5):
        for k in range(5):
            target = torch.zeros(max(j, k) + 1)

            target[j] = target[j] + 1
            target[k] = target[k] - 1

            input = torch.tensor([0.0] * j + [1.0])
            other = torch.tensor([0.0] * k + [1.0])

            torch.testing.assert_close(
                beignet.trim_chebyshev_polynomial_coefficients(
                    beignet.subtract_chebyshev_polynomial(
                        input,
                        other,
                    ),
                    tol=0.000001,
                ),
                beignet.trim_chebyshev_polynomial_coefficients(
                    target,
                    tol=0.000001,
                ),
            )



import beignet
import torch


def test_physicists_hermite_polynomial_vandermonde_3d():
    a, b, x3 = torch.rand(3, 5) * 2 - 1

    coefficients = torch.rand(2, 3, 4)

    output = beignet.physicists_hermite_polynomial_vandermonde_3d(
        a,
        b,
        x3,
        degree=torch.tensor([1, 2, 3]),
    )

    torch.testing.assert_close(
        output @ torch.ravel(coefficients),
        beignet.evaluate_physicists_hermite_polynomial_3d(a, b, x3, coefficients),
    )

    output = beignet.physicists_hermite_polynomial_vandermonde_3d(
        a,
        b,
        x3,
        degree=torch.tensor([1, 2, 3]),
    )

    assert output.shape == (5, 24)



import torch
from torch import Tensor


def compose_quaternion(
    input: Tensor,
    other: Tensor,
    canonical: bool = False,
) -> Tensor:
    r"""
    Compose rotation quaternions.

    Parameters
    ----------
    input : Tensor, shape=(..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    other : Tensor, shape=(..., 4)
        Rotation quaternions. Rotation quaternions are normalized to unit norm.

    canonical : bool, optional
        Whether to map the redundant double cover of rotation space to a unique
        canonical single cover. If `True`, then the rotation quaternion is
        chosen from :math:`{q, -q}` such that the :math:`w` term is positive.
        If the :math:`w` term is :math:`0`, then the rotation quaternion is
        chosen such that the first non-zero term of the :math:`x`, :math:`y`,
        and :math:`z` terms is positive.

    Returns
    -------
    output : Tensor, shape=(..., 4)
        Composed rotation quaternions.
    """
    output = torch.empty(
        [max(input.shape[0], other.shape[0]), 4],
        dtype=input.dtype,
        layout=input.layout,
        device=input.device,
    )

    for j in range(max(input.shape[0], other.shape[0])):
        a = input[j, 0]
        b = input[j, 1]
        c = input[j, 2]
        d = input[j, 3]

        p = other[j, 0]
        q = other[j, 1]
        r = other[j, 2]
        s = other[j, 3]

        t = output[j, 0]
        u = output[j, 1]
        v = output[j, 2]
        w = output[j, 3]

        output[j, 0] = d * p + s * a + b * r - c * q
        output[j, 1] = d * q + s * b + c * p - a * r
        output[j, 2] = d * r + s * c + a * q - b * p
        output[j, 3] = d * s - a * p - b * q - c * r

        x = torch.sqrt(t**2.0 + u**2.0 + v**2.0 + w**2.0)

        if x == 0.0:
            output[j] = torch.nan

        output[j] = output[j] / x

        if canonical:
            if w == 0 and (t == 0 and (u == 0 and v < 0 or u < 0) or t < 0) or w < 0:
                output[j] = -output[j]

    return output



import torch
from torch import Tensor

from ._add_physicists_hermite_polynomial import add_physicists_hermite_polynomial
from ._multiply_physicists_hermite_polynomial_by_x import (
    multiply_physicists_hermite_polynomial_by_x,
)


def polynomial_to_physicists_hermite_polynomial(input: Tensor) -> Tensor:
    input = torch.atleast_1d(input)

    output = torch.zeros_like(input)

    for index in range(0, input.shape[0] - 1 + 1):
        output = multiply_physicists_hermite_polynomial_by_x(output, mode="same")

        output = add_physicists_hermite_polynomial(
            output, input[input.shape[0] - 1 - index]
        )

    return output



from torch import Tensor

from ._quaternion_to_rotation_matrix import (
    quaternion_to_rotation_matrix,
)
from ._rotation_vector_to_quaternion import (
    rotation_vector_to_quaternion,
)


def rotation_vector_to_rotation_matrix(
    input: Tensor,
    degrees: bool | None = False,
) -> Tensor:
    r"""
    Convert rotation vectors to rotation matrices.

    Parameters
    ----------
    input : Tensor, shape=(..., 3)
        Rotation vectors.

    degrees : bool, optional
        If `True`, rotation vector magnitudes are assumed to be in degrees.
        Default, `False`.

    Returns
    -------
    output : Tensor, shape=(..., 3, 3)
        Rotation matrices.
    """
    return quaternion_to_rotation_matrix(
        rotation_vector_to_quaternion(
            input,
            degrees,
        ),
    )



import beignet
import torch


def test_evaluate_physicists_hermite_polynomial_cartesian_2d():
    c1d = torch.tensor([2.5, 1.0, 0.75])
    c2d = torch.einsum("i,j->ij", c1d, c1d)

    x = torch.rand(3, 5) * 2 - 1
    a, b, x3 = x
    y1, y2, y3 = beignet.evaluate_polynomial(x, torch.tensor([1.0, 2.0, 3.0]))

    target = torch.einsum("i,j->ij", y1, y2)
    torch.testing.assert_close(
        beignet.evaluate_physicists_hermite_polynomial_cartesian_2d(
            a,
            b,
            c2d,
        ),
        target,
    )

    z = torch.ones([2, 3])
    assert (
        beignet.evaluate_physicists_hermite_polynomial_cartesian_2d(
            z,
            z,
            c2d,
        ).shape
        == (2, 3) * 2
    )



from pathlib import Path

import pytest
from beignet.datasets import SequenceDataset


class TestSequenceDataset:
    @pytest.mark.parametrize(
        "root",
        [
            "./bar/baz",
            "bar/baz",
            Path("./bar/baz"),
            Path("bar/baz"),
        ],
    )
    def test___init__(self, mocker, root):
        resolve = mocker.patch(
            "pathlib.Path.resolve",
            return_value=Path("/foo/bar/baz"),
        )

        dataset = SequenceDataset(root)

        assert isinstance(dataset.root, Path)

        assert dataset.root == Path("/foo/bar/baz")

        resolve.assert_called_once()



from torch import Tensor

from ._evaluate_physicists_hermite_polynomial import (
    evaluate_physicists_hermite_polynomial,
)


def evaluate_physicists_hermite_polynomial_3d(
    x: Tensor,
    y: Tensor,
    z: Tensor,
    coefficients: Tensor,
) -> Tensor:
    points = [x, y, z]

    if not all(a.shape == points[0].shape for a in points[1:]):
        match len(points):
            case 2:
                raise ValueError
            case 3:
                raise ValueError
            case _:
                raise ValueError

    points = iter(points)

    output = evaluate_physicists_hermite_polynomial(
        next(points),
        coefficients,
    )

    for x in points:
        output = evaluate_physicists_hermite_polynomial(
            x,
            output,
            tensor=False,
        )

    return output
